#include "LSTMNetwork.hpp"

LSTMNetworkImpl::LSTMNetworkImpl(std::map<std::string,std::size_t> nbOutputsPerState, int unknownValueThreshold, std::vector<int> bufferContext, std::vector<int> stackContext, std::vector<std::string> columns, std::vector<int> focusedBufferIndexes, std::vector<int> focusedStackIndexes, std::vector<std::string> focusedColumns, std::vector<int> maxNbElements, int leftWindowRawInput, int rightWindowRawInput, int embeddingsSize, std::vector<std::pair<int, float>> mlpParams, int contextLSTMSize, int focusedLSTMSize, int rawInputLSTMSize, int splitTransLSTMSize, int numLayers, bool bilstm, float lstmDropout, std::vector<std::string> treeEmbeddingColumns, std::vector<int> treeEmbeddingBuffer, std::vector<int> treeEmbeddingStack, std::vector<int> treeEmbeddingNbElems, int treeEmbeddingSize, float embeddingsDropoutValue, float totalInputDropout, bool drop2d)
{
  LSTMImpl::LSTMOptions lstmOptions{true,bilstm,numLayers,lstmDropout,false};
  auto lstmOptionsAll = lstmOptions;
  std::get<4>(lstmOptionsAll) = true;

  int currentOutputSize = embeddingsSize;
  int currentInputSize = 1;

  contextLSTM = register_module("contextLSTM", ContextLSTM(columns, embeddingsSize, contextLSTMSize, bufferContext, stackContext, lstmOptions, unknownValueThreshold));
  contextLSTM->setFirstInputIndex(currentInputSize);
  currentOutputSize += contextLSTM->getOutputSize();
  currentInputSize += contextLSTM->getInputSize();

  if (leftWindowRawInput >= 0 and rightWindowRawInput >= 0)
  {
    rawInputLSTM = register_module("rawInputLSTM", RawInputLSTM(leftWindowRawInput, rightWindowRawInput, embeddingsSize, rawInputLSTMSize, lstmOptionsAll));
    rawInputLSTM->setFirstInputIndex(currentInputSize);
    currentOutputSize += rawInputLSTM->getOutputSize();
    currentInputSize += rawInputLSTM->getInputSize();
  }

  if (!treeEmbeddingColumns.empty())
  {
    treeEmbedding = register_module("treeEmbedding", DepthLayerTreeEmbedding(treeEmbeddingNbElems,embeddingsSize,treeEmbeddingSize,treeEmbeddingColumns,treeEmbeddingBuffer,treeEmbeddingStack,lstmOptions));
    treeEmbedding->setFirstInputIndex(currentInputSize);
    currentOutputSize += treeEmbedding->getOutputSize();
    currentInputSize += treeEmbedding->getInputSize();
  }

  splitTransLSTM = register_module("splitTransLSTM", SplitTransLSTM(Config::maxNbAppliableSplitTransitions, embeddingsSize, splitTransLSTMSize, lstmOptionsAll));
  splitTransLSTM->setFirstInputIndex(currentInputSize);
  currentOutputSize += splitTransLSTM->getOutputSize();
  currentInputSize += splitTransLSTM->getInputSize();

  for (unsigned int i = 0; i < focusedColumns.size(); i++)
  {
    focusedLstms.emplace_back(register_module(fmt::format("LSTM_{}", focusedColumns[i]), FocusedColumnLSTM(focusedBufferIndexes, focusedStackIndexes, focusedColumns[i], maxNbElements[i], embeddingsSize, focusedLSTMSize, lstmOptions)));
    focusedLstms.back()->setFirstInputIndex(currentInputSize);
    currentOutputSize += focusedLstms.back()->getOutputSize();
    currentInputSize += focusedLstms.back()->getInputSize();
  }

  wordEmbeddings = register_module("word_embeddings", torch::nn::Embedding(torch::nn::EmbeddingOptions(maxNbEmbeddings, embeddingsSize)));
  if (drop2d)
    embeddingsDropout2d = register_module("embeddings_dropout2d", torch::nn::Dropout2d(embeddingsDropoutValue));
  else
    embeddingsDropout = register_module("embeddings_dropout", torch::nn::Dropout(embeddingsDropoutValue));
  inputDropout = register_module("input_dropout", torch::nn::Dropout(totalInputDropout));

  mlp = register_module("mlp", MLP(currentOutputSize, mlpParams));

  for (auto & it : nbOutputsPerState)
    outputLayersPerState.emplace(it.first,register_module(fmt::format("output_{}",it.first), torch::nn::Linear(mlp->outputSize(), it.second)));
}

torch::Tensor LSTMNetworkImpl::forward(torch::Tensor input)
{
  if (input.dim() == 1)
    input = input.unsqueeze(0);

  auto embeddings = wordEmbeddings(input);
  if (embeddingsDropout2d.is_empty())
    embeddings = embeddingsDropout(embeddings);
  else
    embeddings = embeddingsDropout2d(embeddings);

  std::vector<torch::Tensor> outputs{embeddings.narrow(1,0,1).squeeze(1)};

  outputs.emplace_back(contextLSTM(embeddings));

  if (!rawInputLSTM.is_empty())
    outputs.emplace_back(rawInputLSTM(embeddings));

  if (!treeEmbedding.is_empty())
    outputs.emplace_back(treeEmbedding(embeddings));

  outputs.emplace_back(splitTransLSTM(embeddings));

  for (auto & lstm : focusedLstms)
    outputs.emplace_back(lstm(embeddings));

  auto totalInput = inputDropout(torch::cat(outputs, 1));

  return outputLayersPerState.at(getState())(mlp(totalInput));
}

std::vector<std::vector<long>> LSTMNetworkImpl::extractContext(Config & config, Dict & dict) const
{
  if (dict.size() >= maxNbEmbeddings)
    util::warning(fmt::format("dict.size()={} > maxNbEmbeddings={}", dict.size(), maxNbEmbeddings));

  std::vector<std::vector<long>> context;
  context.emplace_back();

  context.back().emplace_back(dict.getIndexOrInsert(config.getState()));

  contextLSTM->addToContext(context, dict, config, mustSplitUnknown());

  if (!rawInputLSTM.is_empty())
    rawInputLSTM->addToContext(context, dict, config, mustSplitUnknown());

  if (!treeEmbedding.is_empty())
    treeEmbedding->addToContext(context, dict, config, mustSplitUnknown());

  splitTransLSTM->addToContext(context, dict, config, mustSplitUnknown());

  for (auto & lstm : focusedLstms)
    lstm->addToContext(context, dict, config, mustSplitUnknown());

  if (!mustSplitUnknown() && context.size() > 1)
    util::myThrow(fmt::format("Not in splitUnknown mode, yet context yields multiple variants (size={})", context.size()));

  return context;
}

